/***************************************************************************
    SDL Based Input Handling.

    Populates keys array with user input.
    If porting to a non-SDL platform, you would need to replace this class.

    Copyright Chris White.
    See license.txt for more details.
***************************************************************************/

#include <iostream>
#include <cstdlib> // abs
#include <cstring>
#include "wii/input.hpp"
#include "gccore.h"
#include "wiiuse/wpad.h"
#include "frontend/config.hpp"

Input input;

Input::Input(void)
{
}

Input::~Input(void)
{
}

void Input::init(int pad_id, int* key_config, int* pad_config, int analog, int* axis, int* analog_settings)
{
    this->key_config  = key_config;
    this->pad_config  = pad_config;
    this->analog      = true;//analog;
    this->gamepad     = true;
    this->axis        = axis;
    this->wheel_zone  = analog_settings[0];
    this->wheel_dead  = analog_settings[1];
    this->pedals_dead = analog_settings[2];

    PAD_Init();
    WPAD_Init();
    /*gamepad = SDL_NumJoysticks() > pad_id;

    if (gamepad)
    {
        stick = SDL_JoystickOpen(pad_id);
    }*/

    a_wheel = CENTRE;
}

void Input::close()
{
    //if (gamepad && stick != NULL)
        //SDL_JoystickClose(stick);
}

// Detect whether a key press change has occurred
bool Input::has_pressed(presses p)
{
    return keys[p] && !keys_old[p];
}

// Detect whether key is still pressed
bool Input::is_pressed(presses p)
{
    return keys[p];
}

// Detect whether pressed and clear the press
bool Input::is_pressed_clear(presses p) {
    bool pressed = keys_old2[p];
    keys_old2[p] = false;
    return pressed;
}

template<class T>
constexpr const T& maxabs(const T&a, const T& b) {
  return (std::abs(a)>std::abs(b))?a:b;
}

// copypaste from libogc, wiiuse's joystick handling is disgusting
static void __wii_clampstick(s8 *px,s8 *py,s8 max,s8 xy,s8 min) {
	s32 x,y,signX,signY,d;

	x = *px;
	y = *py;
	if(x>=0) signX = 1;
	else { signX = -1; x = -(x); }

	if(y>=0) signY = 1;
	else { signY = -1; y = -(y); }

	if(x<=min) x = 0;
	else x -= min;

	if(y<=min) y = 0;
	else y -= min;

	if(x!=0 || y!=0) {
		s32 xx,yy,maxy;

		xx = (x * xy);
		yy= (y * xy);
		maxy = (max * xy);
		if(yy<=xx) {
			d = ((x * xy) + (y * (max - xy)));
			if(maxy<d) {
				x = ((x * maxy)/d);
				y = ((y * maxy)/d);
			}
		} else {
			d = ((y * xy) + (x * (max - xy)));
			if(maxy<d) {
				x = ((x * maxy)/d);
				y = ((y * maxy)/d);
			}
		}
		*px = (s8)(x * signX);
		*py = (s8)(y * signY);
	} else
		*px = *py = 0;
}

// Denote that a frame has been done by copying key presses into previous array
void Input::frame_done()
{
    //memcpy(keys_old2, keys_old, sizeof(keys));
    memcpy(keys_old, keys, sizeof(keys));
    PADStatus status[4];
    PAD_Read(status);
    if (status[0].err == PAD_ERR_NO_CONTROLLER) {
      PAD_Reset(PAD_CHAN0_BIT);
      PAD_Read(status);
    }
    PAD_Clamp(status);
    WPAD_SetDataFormat(0, WPAD_FMT_BTNS_ACC);
    WPAD_ScanPads();
    memset(keys,false,sizeof(keys));
    
    s8 wheel =0;
    u8 accel = 0,brake = 0;
    
    if (!status[0].err) {
      keys[UP] |= status[0].button & PAD_BUTTON_UP;
      keys[DOWN] |= status[0].button & PAD_BUTTON_DOWN;
      keys[LEFT] |= status[0].button & PAD_BUTTON_LEFT;
      keys[RIGHT] |= status[0].button & PAD_BUTTON_RIGHT;
      keys[ACCEL] |= status[0].button & PAD_BUTTON_A;
      keys[BRAKE] |= status[0].button & PAD_BUTTON_B;
      keys[VIEWPOINT] |= status[0].button & PAD_BUTTON_Y;
      keys[GEAR1] |= status[0].button & PAD_TRIGGER_Z;
      keys[START] |= (status[0].button & PAD_BUTTON_START);
      
      wheel = maxabs<s8>(wheel,((status[0].stickX*1000)/1285));
      u8 rtrig = ((status[0].triggerR*1000)/1171);
      u8 ltrig = ((status[0].triggerL*1000)/1171);
      accel = std::max<u8>(accel,rtrig?0x30+rtrig:0);
      brake = std::max<u8>(brake,ltrig?0x30+ltrig:0);
    }
    
    auto wiimote1 = WPAD_ButtonsHeld(0);
    keys[MENU] |= wiimote1 & WPAD_BUTTON_HOME;
    keys[PAUSE] |= wiimote1 & WPAD_BUTTON_MINUS;
    keys[START] |= wiimote1 & WPAD_BUTTON_PLUS;
    
    WPADData* data = WPAD_Data(0);
    
    struct expansion_t &exp = data->exp;
    
    if (exp.type == WPAD_EXP_CLASSIC) {
      keys[UP] |= wiimote1 & WPAD_CLASSIC_BUTTON_UP;
      keys[DOWN] |= wiimote1 & WPAD_CLASSIC_BUTTON_DOWN;
      keys[LEFT] |= wiimote1 & WPAD_CLASSIC_BUTTON_LEFT;
      keys[RIGHT] |= wiimote1 & WPAD_CLASSIC_BUTTON_RIGHT;
      keys[ACCEL] |= wiimote1 & WPAD_CLASSIC_BUTTON_A;
      keys[BRAKE] |= wiimote1 & WPAD_CLASSIC_BUTTON_B;
      keys[GEAR1] |= wiimote1 & WPAD_CLASSIC_BUTTON_FULL_R;
      keys[VIEWPOINT] |= wiimote1 & WPAD_CLASSIC_BUTTON_Y;
      keys[START] |= wiimote1 & WPAD_CLASSIC_BUTTON_PLUS;
      keys[PAUSE] |= wiimote1 & WPAD_CLASSIC_BUTTON_MINUS;
      keys[MENU] |= wiimote1 & WPAD_CLASSIC_BUTTON_HOME;
      
      s8 stickx = 4*(exp.classic.ljs.pos.x - exp.classic.ljs.center.x);
      s8 sticky = 4*(exp.classic.ljs.pos.y - exp.classic.ljs.center.y);
      __wii_clampstick(&stickx,&sticky,72,40,15);
      wheel = maxabs<s8>(wheel,((stickx*1000)/1285));
    }
    if (exp.type == WPAD_EXP_NUNCHUK) {
      keys[UP] |= wiimote1 & WPAD_BUTTON_UP;
      keys[DOWN] |= wiimote1 & WPAD_BUTTON_DOWN;
      keys[LEFT] |= wiimote1 & WPAD_BUTTON_LEFT;
      keys[RIGHT] |= wiimote1 & WPAD_BUTTON_RIGHT;
      keys[ACCEL] |= wiimote1 & WPAD_BUTTON_A;
      keys[BRAKE] |= wiimote1 & WPAD_BUTTON_B;
      keys[GEAR1] |= wiimote1 & WPAD_NUNCHUK_BUTTON_Z;
      keys[VIEWPOINT] |= wiimote1 & WPAD_BUTTON_1;
      
      s8 stickx = exp.nunchuk.js.pos.x - exp.nunchuk.js.center.x;
      s8 sticky = exp.nunchuk.js.pos.y - exp.nunchuk.js.center.y;
      __wii_clampstick(&stickx,&sticky,72,40,15);
      wheel = maxabs<s8>(wheel,((stickx*1000)/1285));
    }
    if (exp.type == WPAD_EXP_NONE) {
      keys[UP] |= wiimote1 & WPAD_BUTTON_RIGHT;
      keys[DOWN] |= wiimote1 & WPAD_BUTTON_LEFT;
      keys[LEFT] |= wiimote1 & WPAD_BUTTON_UP;
      keys[RIGHT] |= wiimote1 & WPAD_BUTTON_DOWN;
      keys[ACCEL] |= wiimote1 & WPAD_BUTTON_2;
      keys[BRAKE] |= wiimote1 & WPAD_BUTTON_1;
      keys[GEAR1] |= wiimote1 & WPAD_BUTTON_B;
      keys[VIEWPOINT] |= wiimote1 & WPAD_BUTTON_A;
      
      if (config.controls.wiiwheel) {
        f32 motion = std::max(std::min(data->orient.pitch/-90.f,1.f),-1.f);
        wheel = maxabs<s8>(wheel,s8(motion*57));
      }
    }
    
    a_wheel = 0x80 + wheel;
    a_accel = accel;
    a_brake = brake;
    
    for (int i=0;i<sizeof(keys);i++) {
      if (keys[i] && !keys_old[i]) keys_old2[i] = true;
    }
}

/*void Input::handle_key_down(SDL_keysym* keysym)
{
    key_press = keysym->sym;
    handle_key(key_press, true);
}

void Input::handle_key_up(SDL_keysym* keysym)
{
    handle_key(keysym->sym, false);
}*/
void Input::handle_key(const int key, const bool is_pressed)
{
    // Redefinable Key Input
    if (key == key_config[0])
        keys[UP] = is_pressed;

    else if (key == key_config[1])
        keys[DOWN] = is_pressed;

    else if (key == key_config[2])
        keys[LEFT] = is_pressed;

    else if (key == key_config[3])
        keys[RIGHT] = is_pressed;

    if (key == key_config[4])
        keys[ACCEL] = is_pressed;

    if (key == key_config[5])
        keys[BRAKE] = is_pressed;

    if (key == key_config[6])
        keys[GEAR1] = is_pressed;

    if (key == key_config[7])
        keys[GEAR2] = is_pressed;

    if (key == key_config[8])
        keys[START] = is_pressed;

    if (key == key_config[9])
        keys[COIN] = is_pressed;

    if (key == key_config[10])
        keys[MENU] = is_pressed;

    if (key == key_config[11])
        keys[VIEWPOINT] = is_pressed;

    // Function keys are not redefinable
   /* switch (key)
    {
        case SDLK_F1:
            keys[PAUSE] = is_pressed;
            break;

        case SDLK_F2:
            keys[STEP] = is_pressed;
            break;

        case SDLK_F3:
            keys[TIMER] = is_pressed;
            break;

        case SDLK_F5:
            keys[MENU] = is_pressed;
            break;
    }*/
}

/*void Input::handle_joy_axis(SDL_JoyAxisEvent* evt)
{
    int16_t value = evt->value;

    // Digital Controls
    if (!analog)
    {
        // X-Axis
        if (evt->axis == 0)
        {
            // Neural
            if ( (value > -DIGITAL_DEAD ) && (value < DIGITAL_DEAD ) )
            {
                keys[LEFT]  = false;
                keys[RIGHT] = false;
            }
            else if (value < 0)
            {
                keys[LEFT] = true;
            }
            else if (value > 0)
            {
                keys[RIGHT] = true;
            }
        }
        // Y-Axis
        else if (evt->axis == 1)
        {
            // Neural
            if ( (value > -DIGITAL_DEAD ) && (value < DIGITAL_DEAD ) )
            {
                keys[UP]  = false;
                keys[DOWN] = false;
            }
            else if (value < 0)
            {
                keys[UP] = true;
            }
            else if (value > 0)
            {
                keys[DOWN] = true;
            }
        }
    }
    // Analog Controls
    else
    {
        //std::cout << "Axis: " << (int) evt->axis << " Value: " << (int) evt->value << std::endl;

        // Steering
        // OutRun requires values between 0x48 and 0xb8.
        if (evt->axis == axis[0])
        {
            int percentage_adjust = ((wheel_zone) << 8) / 100;         
            int adjusted = value + ((value * percentage_adjust) >> 8);
            
            // Make 0 hard left, and 0x80 centre value.
            adjusted = ((adjusted + (1 << 15)) >> 9);
            adjusted += 0x40; // Centre

            if (adjusted < 0x40)
                adjusted = 0x40;
            else if (adjusted > 0xC0)
                adjusted = 0xC0;

            // Remove Dead Zone
            if (wheel_dead)
            {
                if (std::abs(CENTRE - adjusted) <= wheel_dead)
                    adjusted = CENTRE;
            }

            //std::cout << "wheel zone : " << wheel_zone << " : " << std::hex << " : " << (int) adjusted << std::endl;
            a_wheel = adjusted;
        }
        // Accelerator and Brake [Combined Axis]
        else if (axis[1] == axis[2] && (evt->axis == axis[1] || evt->axis == axis[2]))
        {
            // Accelerator
            if (value < -pedals_dead)
            {
                // Scale input to be in the range of 0 to 0x7F
                value = -value;
                int adjusted = value / 258;
                adjusted += (adjusted >> 2);
                a_accel = adjusted;
            }
            // Brake
            else if (value > pedals_dead)
            {
                // Scale input to be in the range of 0 to 0x7F
                int adjusted = value / 258;
                adjusted += (adjusted >> 2);
                a_brake = adjusted;
            }
            else
            {
                a_accel = 0;
                a_brake = 0;
            }
        }
        // Accelerator [Single Axis]
        else if (evt->axis == axis[1])
        {
            // Scale input to be in the range of 0 to 0x7F
            int adjusted = 0x7F - ((value + (1 << 15)) >> 9);           
            adjusted += (adjusted >> 2);
            a_accel = adjusted;
        }
        // Brake [Single Axis]
        else if (evt->axis == axis[2])
        {
            // Scale input to be in the range of 0 to 0x7F
            int adjusted = 0x7F - ((value + (1 << 15)) >> 9);
            adjusted += (adjusted >> 2);
            a_brake = adjusted;
        }
    }
}

void Input::handle_joy_down(SDL_JoyButtonEvent* evt)
{
    // Latch joystick button presses for redefines
    joy_button = evt->button;
    handle_joy(evt->button, true);
}

void Input::handle_joy_up(SDL_JoyButtonEvent* evt)
{
    handle_joy(evt->button, false);
}*/

void Input::handle_joy(const uint8_t button, const bool is_pressed)
{
    if (button == pad_config[0])
        keys[ACCEL] = is_pressed;

    if (button == pad_config[1])
        keys[BRAKE] = is_pressed;

    if (button == pad_config[2])
        keys[GEAR1] = is_pressed;

    if (button == pad_config[3])
        keys[GEAR2] = is_pressed;

    if (button == pad_config[4])
        keys[START] = is_pressed;

    if (button == pad_config[5])
        keys[COIN] = is_pressed;

    if (button == pad_config[6])
        keys[MENU] = is_pressed;

    if (button == pad_config[7])
        keys[VIEWPOINT] = is_pressed;
}